#include "gui.hpp"
#include <iostream>
#include <chrono>
#include <numeric>
#include <functional>
#include "components.hpp"
#include <numbers>
#include "systems.hpp"
#include "events.hpp"

namespace gui {
  using namespace components;

  FSM::FSM() :
    $window(sf::VideoMode({SCREEN_WIDTH, SCREEN_HEIGHT}), "Zed's Raycaster Thing"),
    $renderer($window),
    $level($levels.current()),
    $map_view($level),
    $combat_view($level),
    $status_view($level),
    $font{FONT_FILE_NAME},
    $text{$font},
    $rayview($textures, RAY_VIEW_WIDTH, RAY_VIEW_HEIGHT)
    {
      $window.setVerticalSyncEnabled(VSYNC);
      $window.setFramerateLimit(FRAME_LIMIT);
      $text.setPosition({RAY_VIEW_X,RAY_VIEW_Y});
      $text.setCharacterSize(20);
      $text.setFillColor(ColorValue::LIGHT_DARK);
      $textures.load_tiles();
      $textures.load_sprites();
    }

  void FSM::event(Event ev) {
    switch($state) {
      FSM_STATE(State, START, ev);
      FSM_STATE(State, MOVING, ev);
      FSM_STATE(State, ATTACKING, ev);
      FSM_STATE(State, MAPPING, ev);
      FSM_STATE(State, ROTATING, ev);
      FSM_STATE(State, IDLE, ev);
      FSM_STATE(State, IN_COMBAT, ev);
      FSM_STATE(State, COMBAT_ROTATE, ev);
      FSM_STATE(State, END, ev);
    }
  }

  void FSM::START(Event ) {
    generate_map();
    $level.world->set_the<Debug>({});
    $rayview.init_shaders();
    $rayview.set_position(RAY_VIEW_X, RAY_VIEW_Y);
    $rayview.position_camera($player.x + 0.5, $player.y + 0.5);

    $combat_view.render($textures);

    $status_view.render($textures);
    $status_view.log("Welcome to the game!");

    $renderer.init_terminal();
    $map_view.create_render();
    $map_view.resize_canvas();
    $renderer.resize_grid(MAX_FONT_SIZE, $map_view);

    run_systems();
    state(State::IDLE);
  }

  void FSM::MAPPING(Event ev) {
    using enum Event;
    // BUG: can't close window when in mapping
    switch(ev) {
      case MAP_OPEN:
        state(State::IDLE);
        break;
      case CLOSE:
        state(State::IDLE);
        break;
      case TICK:
        break;
      default:
        dbc::log("invalid event sent to MAPPING");
    }
  }

  void FSM::MOVING(Event ) {
    if($camera.play_move($rayview)) {
      System::plan_motion(*$level.world, {size_t($camera.target_x), size_t($camera.target_y)});
      run_systems();
      state(State::IDLE);
    }
  }

  void FSM::ATTACKING(Event ev) {
    using enum Event;
    switch(ev) {
      case TICK:
        $rotation_count++;

        // just do 10 ticks
        if($rotation_count % 10 == 0) {
          System::combat($level);
          run_systems();
          $rotation = -10.0f;
          state(State::IN_COMBAT);
        }
        break;
      case STOP_COMBAT:
        dbc::log("Exiting ATTACKING STATE");
        state(State::IDLE);
        break;
      case ATTACK:
        // ignore these since they're just from SFML not having discrete events
        break;
      default:
        dbc::log(fmt::format("In ATTACKING state, unhandled event {}", (int)ev));
    }
  }

  void FSM::ROTATING(Event ) {
    if($camera.play_rotate($rayview)) {
      state(State::IDLE);
    }
  }

  void FSM::COMBAT_ROTATE(Event ) {
    if($camera.play_rotate($rayview)) {
      state(State::IN_COMBAT);
    }
  }

  void FSM::IDLE(Event ev) {
    using enum Event;

    switch(ev) {
      case QUIT:
        $window.close();
        state(State::END);
        return; // done
      case MOVE_FORWARD:
        try_move(1, false);
        break;
      case MOVE_BACK:
        try_move(-1, false);
        break;
      case MOVE_LEFT:
        try_move(1, true);
        break;
      case MOVE_RIGHT:
        try_move(-1, true);
        break;
      case ROTATE_LEFT:
        $camera.plan_rotate($rayview, 1);
        state(State::ROTATING);
        break;
      case ROTATE_RIGHT:
        $camera.plan_rotate($rayview, -1);
        state(State::ROTATING);
        break;
      case MAP_OPEN:
        $renderer.resize_grid(MAX_FONT_SIZE, $map_view);
        $map_view.resize_canvas();
        state(State::MAPPING);
        break;
      case ATTACK:
        state(State::ATTACKING);
        break;
      case START_COMBAT:
        state(State::IN_COMBAT);
        break;
      case CLOSE:
        dbc::log("Nothing to close.");
        break;
      case STOP_COMBAT:
      case TICK:
        // do nothing
        break;
      default:
        dbc::sentinel("unhandled event in IDLE");
    }
  }

  void FSM::IN_COMBAT(Event ev) {
    using enum Event;

    switch(ev) {
      case ATTACK:
        $status_view.log("You attack!");
        $rotation = -30.0f;
        state(State::ATTACKING);
        break;
      case ROTATE_LEFT:
        $camera.plan_rotate($rayview, 1);
        state(State::COMBAT_ROTATE);
        break;
      case ROTATE_RIGHT:
        $camera.plan_rotate($rayview, -1);
        state(State::COMBAT_ROTATE);
        break;
      case STOP_COMBAT:
        state(State::IDLE);
        break;
      case QUIT:
        $window.close();
        state(State::END);
        return;
      default:
        break;
    }
  }

  void FSM::try_move(int dir, bool strafe) {
    using enum State;
    // prevent moving into occupied space
    Point move_to = $camera.plan_move($rayview, dir, strafe);

    if($level.map->can_move(move_to) && !$level.collision->occupied(move_to)) {
      state(MOVING);
    } else {
      state(IDLE);
      $camera.abort_plan($rayview);
    }
  }

  void FSM::END(Event ev) {
    fmt::println("END: received event after done: {}", int(ev));
  }

  void FSM::keyboard() {
    while(const auto keyev = $window.pollEvent()) {
      if(keyev->is<sf::Event::Closed>()) {
        event(Event::QUIT);
      }

      if(const auto* key = keyev->getIf<sf::Event::KeyPressed>()) {
        using KEY = sf::Keyboard::Scan;
        switch(key->scancode) {
          case KEY::W:
            event(Event::MOVE_FORWARD);
            break;
          case KEY::S:
            event(Event::MOVE_BACK);
            break;
          case KEY::Q:
            event(Event::ROTATE_LEFT);
            break;
          case KEY::E:
            event(Event::ROTATE_RIGHT);
            break;
          case KEY::D:
            event(Event::MOVE_RIGHT);
            break;
          case KEY::A:
            event(Event::MOVE_LEFT);
            break;
          case KEY::R:
            $stats.reset();
            break;
          case KEY::M:
            event(Event::MAP_OPEN);
            break;
          case KEY::Escape:
            event(Event::CLOSE);
            break;
          case KEY::P:
            debug();
          default:
            break; // ignored
        }
      }
    }
  }

  void FSM::debug() {
    auto& debug = $level.world->get_the<Debug>();
    debug.FPS = !debug.FPS;
    debug.PATHS = !debug.PATHS;
    auto player = $level.world->get_the<Player>();
    auto& player_combat = $level.world->get<Combat>(player.entity);
    player_combat.hp = player_combat.max_hp;
    $combat_view.set_damage(float(player_combat.hp) / float(player_combat.max_hp));
  }

  void FSM::draw_weapon() {
    return;
    auto weapon = $rayview.$textures.sword.sprite;
    weapon->setPosition({SCREEN_WIDTH/2, SCREEN_HEIGHT/2});
    weapon->setRotation(sf::degrees($rotation));
    $window.draw(*weapon);
  }

  void FSM::draw_stats() {
    auto player = $level.world->get_the<Player>();
    auto player_combat = $level.world->get<Combat>(player.entity);

    $text.setString(
        fmt::format("FPS\n"
          "HP: {}\n"
          "mean:{:>8.5}\n"
          "sdev: {:>8.5}\n"
          "min: {:>8.5}\n"
          "max: {:>8.5}\n"
          "count:{:<10}\n\n"
          "VSync? {}\n"
          "FR Limit: {}\n"
          "Debug? {}\n\n",
          player_combat.hp, $stats.mean(), $stats.stddev(), $stats.min,
          $stats.max, $stats.n, VSYNC,
          FRAME_LIMIT, DEBUG_BUILD));

    $window.draw($text);
  }

  void FSM::draw_blood() {
    auto player = $level.world->get_the<Player>();
    auto player_combat = $level.world->get<Combat>(player.entity);

    if(float(player_combat.hp) / float(player_combat.max_hp) < 0.5) {
      auto blood = $textures.sprite_textures.at("blood_splatter").sprite;
      blood->setPosition({RAY_VIEW_X,0});
      blood->setScale({3.0, 3.0});
      $window.draw(*blood);
    }
  }

  void FSM::draw_gui() {
    $status_view.draw($window);
    $combat_view.draw($window);

    auto debug = $level.world->get_the<Debug>();
    if(debug.FPS) draw_stats();
  }

  void FSM::render() {
    if(in_state(State::MAPPING)) {
      $window.clear();
      $map_view.render();
      $renderer.draw($map_view);
    } else {
      auto start = std::chrono::high_resolution_clock::now();
      $rayview.draw($window);
      auto end = std::chrono::high_resolution_clock::now();
      auto elapsed = std::chrono::duration<double>(end - start);
      $stats.sample(1/elapsed.count());

      draw_gui();
      draw_blood();
      draw_weapon();
    }

    $window.display();
  }

  void FSM::mouse() {
    if(sf::Mouse::isButtonPressed(sf::Mouse::Button::Left)) {
      sf::Vector2f pos = $window.mapPixelToCoords(sf::Mouse::getPosition($window));
      $combat_view.$gui.mouse(pos.x, pos.y);
      $status_view.$gui.mouse(pos.x, pos.y);
    }
  }

  void FSM::generate_map() {
    // ZED: this should eventually go away now that level manager is in play
    auto& player = $level.world->get_the<Player>();
    auto& player_position = $level.world->get<Position>(player.entity);
    $player = player_position.location;
    $rayview.set_level($level);
  }

  void FSM::run_systems() {
    System::enemy_pathing($level);
    System::collision($level);
    System::motion($level);
    System::lighting($level);
    System::death($level);
  }

  bool FSM::active() {
    return !in_state(State::END);
  }

  void FSM::handle_world_events() {
    using eGUI = Events::GUI;
    auto& world = *$level.world;

    while(world.has_event<eGUI>()) {
      auto [evt, entity, data] = world.recv<eGUI>();

      switch(evt) {
        case eGUI::COMBAT: {
            auto &damage = std::any_cast<Events::Combat&>(data);
            auto enemy_combat = world.get<Combat>(entity);

            if(damage.enemy_did > 0) {
              $status_view.log(fmt::format("Enemy HIT YOU for {} damage!", damage.enemy_did));
              $status_view.log(fmt::format("-- Enemy has {} HP left.", enemy_combat.hp));
              auto player = world.get_the<Player>();
              auto player_combat = world.get<Combat>(player.entity);
              $combat_view.set_damage(float(player_combat.hp) / float(player_combat.max_hp));
            } else {
              $status_view.log("Enemy MISSED YOU.");
            }

            if(damage.player_did > 0) {
              $status_view.log(fmt::format("You HIT enemy for {} damage!", damage.player_did));
            } else {
              $status_view.log("You MISSED the enemy.");
            }
          }
          break;
        case eGUI::COMBAT_START:
          event(Event::START_COMBAT);
          break;
        case eGUI::NO_NEIGHBORS:
          event(Event::STOP_COMBAT);
          break;
        case eGUI::LOOT: {
            // auto &item = std::any_cast<InventoryItem&>(data);
            // $status_view.log(fmt::format("You picked up a {}.",
            //      std::string(item.data["name"])));
            $status_view.log("You picked up an item.");
          }
          break;
        case eGUI::ATTACK:
          $rotation = 0;
          event(Event::ATTACK);
          break;
        case eGUI::NOOP:
          $status_view.log(fmt::format("NOOP EVENT! {},{}", evt, entity));
          break;
        default:
          $status_view.log(fmt::format("INVALID EVENT! {},{}", evt, entity));
      }
    }
  }
}
